from typing import List, Optional, Union

from django.contrib.auth import models as auth_models
from django.db.models.manager import EmptyManager
from django.utils.functional import cached_property


class MicroserviceUser:
    is_active = True

    _groups = EmptyManager(auth_models.Group)
    _user_permissions = EmptyManager(auth_models.Permission)

    def __str__(self) -> str:
        return f"{self.id}"

    @cached_property
    def id(self) -> Union[int, str]:
        return -1

    @cached_property
    def pk(self) -> Union[int, str]:
        return self.id

    @cached_property
    def username(self) -> str:
        return 'microservice'

    @cached_property
    def is_staff(self) -> bool:
        return True

    @cached_property
    def is_superuser(self) -> bool:
        return True

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MicroserviceUser):
            return NotImplemented
        return self.id == other.id

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self.id)

    def save(self) -> None:
        raise NotImplementedError("Microservice users have no DB representation")

    def delete(self) -> None:
        raise NotImplementedError("Microservice users have no DB representation")

    def set_password(self, raw_password: str) -> None:
        raise NotImplementedError("Microservice users have no DB representation")

    def check_password(self, raw_password: str) -> None:
        raise NotImplementedError("Microservice users have no DB representation")

    @property
    def groups(self) -> auth_models.Group:
        return self._groups

    @property
    def user_permissions(self) -> auth_models.Permission:
        return self._user_permissions

    def get_group_permissions(self, obj: Optional[object] = None) -> set:
        return set()

    def get_all_permissions(self, obj: Optional[object] = None) -> set:
        return set()

    def has_perm(self, perm: str, obj: Optional[object] = None) -> bool:
        return False

    def has_perms(self, perm_list: List[str], obj: Optional[object] = None) -> bool:
        return False

    def has_module_perms(self, module: str) -> bool:
        return False

    @property
    def is_anonymous(self) -> bool:
        return False

    @property
    def is_authenticated(self) -> bool:
        return True

    def get_username(self) -> str:
        return self.username



import os

from requests import get


class BaseMicroserviceSetting:
    url: str = None
    token: str = None


class UserServiceSetting(BaseMicroserviceSetting):
    url = os.environ.get("USER_SERVICE_URL")


class UserService:
    base_url = UserServiceSetting.url

    @staticmethod
    def url(path):
        return UserService.base_url + path

    @staticmethod
    def get_users(ids: list[str]) -> list[dict]:
        return get(UserService.url('/api/users/all/'), {'ids': ','.join(ids)}).json()



import os

import requests
from django.utils.translation import get_language


def translate(key: str) -> str:
    base_url = os.environ.get('TRANSLATION_SERVICE_URL')
    lang = get_language()
    if lang == 'uz-cyr':
        lang = 'crl'
    return requests.get(f'{base_url}/api/translations/{key}', params={
        'lang': lang
    }).text



from rest_framework_simplejwt.models import TokenUser
from rest_framework_simplejwt.tokens import Token


class JWTUser(TokenUser):

    def __init__(self, token: "Token"):
        super().__init__(token)
        self.phone = self.token.get("phone")
        self._groups = self.token.get("groups")



import importlib
import mimetypes
import os
import uuid

from django.conf import settings
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.http import HttpResponse
from django_filters.rest_framework.backends import DjangoFilterBackend
from rest_framework import viewsets, filters, status
from rest_framework.parsers import MultiPartParser
from rest_framework.response import Response

from ..data_processors import XlsProcessor
from ..filters import FileFilterSet
from ..models import ImportFile, TemporaryData
from ..serializers.file import (
    FileDownloadSerializer,
    ImportFileSerializer,
    FileUploadSerializer,
    FileConfirmSerializer
)
from ..services import minio_service
from ...api import pagination
from ...api.base import BaseAPIView, BaseRequest
from ...services import UserService

data_processors = []
for processor_class in settings.DATA_PROCESSORS_CLASSES:
    module_name, class_name = processor_class.rsplit('.', 1)
    module = importlib.import_module(module_name)
    data_processors.append(getattr(module, class_name))


class ImportFileViewSet(viewsets.GenericViewSet, viewsets.generics.ListAPIView):
    queryset = ImportFile.objects.all()
    serializer_class = ImportFileSerializer
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend)
    filterset_class = FileFilterSet
    pagination_class = pagination.SmallResultsSetPagination

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        queryset = self.paginate_queryset(queryset)

        files = list(queryset)

        created_bys = []
        for file in files:
            created_bys.append(str(file.created_by))

        users = UserService.get_users(created_bys)
        users = {u['id']: u for u in users}

        for file in files:
            creator = users.get(str(file.created_by), {})

            file.created_by = {
                'id': creator.get('id'),
                'username': creator.get('username'),
                'firstName': creator.get('firstName'),
                'surName': creator.get('surName'),
                'midName': creator.get('midName'),
            }

        serializer = self.get_serializer(files, many=True)
        return self.get_paginated_response(serializer.data)


class FileDownloadViewSet(viewsets.generics.RetrieveAPIView):
    queryset = ImportFile.objects.all()
    serializer_class = FileDownloadSerializer

    def retrieve(self, request, *args, **kwargs):
        obj = self.get_object()
        file_bytes = minio_service.get_object('liquid-hydrocarbons', obj.store_name)
        content_type, _ = mimetypes.guess_type(obj.real_name, strict=False)
        response = HttpResponse(file_bytes, content_type=content_type or 'application/octet-stream')
        response['Content-Length'] = len(file_bytes)
        response['Content-Disposition'] = 'attachment; filename="%s"' % obj.real_name
        return response


class FileUploadViewSet(BaseAPIView, viewsets.generics.CreateAPIView):
    serializer_class = FileUploadSerializer
    parser_classes = (MultiPartParser,)

    def create(self, request: BaseRequest, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        file: InMemoryUploadedFile = serializer.validated_data.get('file')
        xls_processor = XlsProcessor(file.read())
        errors = xls_processor.validate_many([data_processor() for data_processor in data_processors])

        if errors:
            return Response({
                'sheets': errors
            }, status=status.HTTP_400_BAD_REQUEST)

        file.seek(0)

        random_str = uuid.uuid4().hex
        date = serializer.validated_data.get('date').strftime('%Y-%m-%d')
        file_extension = os.path.splitext(file.name)[1]
        store_name = f'{date}--{random_str}{file_extension}'

        minio_service.put_object('liquid-hydrocarbons', store_name, file)

        import_file = ImportFile(
            date=serializer.validated_data.get('date'),
            real_name=file.name,
            store_name=store_name,
            size=file.size,
            status=ImportFile.Status.DRAFT,
            active=False,
            created_by=self.request.user.pk,
        )
        import_file.save()

        data = {}
        for data_processor in data_processors:
            data[data_processor.sheet_index] = xls_processor.get_data(data_processor())

        temporary_data = TemporaryData(
            file=import_file,
            data=data,
            created_by=self.request.user.pk
        )
        temporary_data.save()

        return Response(
            data={
                'success': True,
                'file_id': import_file.pk
            },
            status=status.HTTP_201_CREATED
        )


class FileConfirmViewSet(BaseAPIView, viewsets.generics.UpdateAPIView):
    serializer_class = FileConfirmSerializer
    http_method_names = ['patch']
    queryset = ImportFile.objects.filter(status=ImportFile.Status.DRAFT, active=False).all()

    def perform_update(self, serializer):
        model: ImportFile = self.get_object()
        temp = TemporaryData.objects.filter(file=model).get()
        data = temp.data
        for data_processor in data_processors:
            data_processor().write_db(
                data.get(str(data_processor.sheet_index)),
                model,
                self.request.user.pk,
            )
        temp.delete()
        ImportFile.objects.filter(active=True, date=model.date).update(active=False)
        serializer.save(status=ImportFile.Status.SUCCESS, active=True)



from copy import deepcopy


def deep_map(data: dict | list, func_cond, func_map, in_place=True):
    if not in_place:
        data = deepcopy(data)

    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, (list, dict)):
                deep_map(value, func_cond, func_map, True)
            elif func_cond(value):
                data[key] = func_map(value)
    elif isinstance(data, list):
        for index, value in enumerate(data):
            if isinstance(value, (list, dict)):
                deep_map(value, func_cond, func_map, True)
            elif func_cond(value):
                data[index] = func_map(value)

    return data


def null_to_zero(data: dict | list, in_place=True):
    return deep_map(data, lambda value: value is None, lambda _: 0, in_place)


def deep_round(data: dict | list, ndigits: int, in_place=True):
    return deep_map(data, lambda value: isinstance(value, float), lambda value: round(value, ndigits), in_place)


class EData:
    def __init__(self, data):
        self.data = data

    def null_to_zero(self):
        null_to_zero(self.data)
        return self

    def round(self, ndigits: int):
        deep_round(self.data, ndigits)
        return self

    def map(self, func_cond, func_map):
        deep_map(self.data, func_cond, func_map)
        return self



from django.core.validators import FileExtensionValidator
from rest_framework import serializers

from ..models import ImportFile


class ImportFileSerializer(serializers.ModelSerializer):
    created_by = serializers.DictField()

    class Meta:
        model = ImportFile
        fields = '__all__'


class FileDownloadSerializer(serializers.Serializer):
    file = serializers.FileField()


class FileUploadSerializer(serializers.Serializer):
    file = serializers.FileField(validators=[FileExtensionValidator(allowed_extensions=['xls', 'xlsx'])])
    date = serializers.DateField()


class FileConfirmSerializer(serializers.ModelSerializer):
    class Meta:
        model = ImportFile
        fields = []



from django.db.models import Manager
from django.utils.translation import get_language, activate
from rest_framework import generics


class TranslatedListView(generics.ListAPIView):
    manager: Manager

    def get_queryset(self):
        self.update_lang()
        return self.manager.all()

    def update_lang(self):
        activate(self.get_language())

    def get_language(self):
        return self.request.headers.get('accept-language') or get_language()



from django_q.models import Task
from drf_spectacular.utils import extend_schema
from rest_framework import viewsets

from energy_base.q_api.filters import TaskFilterSet
from energy_base.api.permissions import IsSuperUser
from energy_base.q_api.serializers import TaskSerializer


@extend_schema(tags=['admin/tasks'])
class TaskViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = Task.objects.all()
    serializer_class = TaskSerializer
    permission_classes = [IsSuperUser]
    filterset_class = TaskFilterSet




from .file import ImportFileSerializer, FileDownloadSerializer, FileUploadSerializer, FileConfirmSerializer
from .file_upload_status import FileUploadStatusSerializer, FileUploadStatusFilterSerializer

__all__ = [
    'ImportFileSerializer',
    'FileUploadSerializer',
    'FileConfirmSerializer',
    'FileUploadStatusSerializer',
    'FileDownloadSerializer',
    'FileUploadStatusFilterSerializer',
]



from typing import Callable, Any

import requests
import urllib3
from requests import Response

from energy_base.utils.response_processors import call_processor

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class BaseServiceAPI:
    name: str = None
    base_url: str = None
    default_response_processor: Callable[[Response], Any] = None

    def request(self, method, path, data=None, json=None, params=None, headers=None,
                authenticator: Callable[[dict], dict] | None = 'default',
                response_processor: Callable[[Response], Any] | None = 'default',
                **kwargs):
        """
        Make an HTTP request using the requests' library.

        Args:
            method: The HTTP method (GET, POST, PUT, DELETE, etc.).
            path: The API endpoint path.
            data: Dictionary or bytes to be sent in the body of the request (for form data).
            json: JSON serializable data to be sent in the body of the request (for JSON data).
            params: Query parameters as a dictionary.
            headers: Headers to be included in the request.
            authenticator: Optional authenticator function to handle authentication with headers.
                This function should modify the headers for authentication purposes.
                Example authenticator function:
                def my_authenticator(headers):
                    headers['Authorization'] = 'Bearer YOUR_ACCESS_TOKEN'
                    return headers

            response_processor: Optional function to process the response data.
                This function should take the requests.Response object as input and return processed data.
                Example response processor function:
                def process_response(response):
                    return response.json()
        Kwargs:
            auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.

        Returns:
            The response data, processed if a response processor is provided.
        """

        if headers is None:
            headers = {}

        if callable(authenticator):
            headers = authenticator(headers)

        if authenticator == 'default' and hasattr(self, 'authenticate') and callable(self.authenticate):
            headers = self.authenticate(headers)

        response = requests.request(
            method=method,
            url=self.base_url + path,
            data=data,
            json=json,
            params=params,
            headers=headers,
            verify=False,
            **kwargs
        )

        if response_processor == 'default':
            if self.__class__.default_response_processor is not None:
                return self.__class__.default_response_processor(response)
            return response

        if response_processor is not None:
            return response_processor(response)

        return response

    def call(self, method, path, data=None, json=None, params=None, headers=None):
        data = self.request(
            method=method,
            path=path,
            data=data,
            json=json,
            params=params,
            headers=headers,
            response_processor=call_processor
        )
        return {
            'path': self.base_url + path,
            'method': method,
            'status': data['status'],
            'reason': data['reason'],
            'request': params if method == 'get' else json,
            'response': data['response'],
        }

    def authenticate(self, headers: dict) -> dict:
        """
        Example authenticate method:
            authenticate(self, headers):
                headers['Authorization'] = 'Bearer YOUR_ACCESS_TOKEN'
                return headers
        :param headers:
        :return:
        """
        return headers

    def get(self, path, data=None, json=None, params=None, headers=None,
            authenticator: Callable[[dict], dict] | None = 'default',
            response_processor: Callable[[Response], Any] | None = 'default'):
        return self.request('get', path, data, json, params, headers, authenticator, response_processor)

    def post(self, path, data=None, json=None, params=None, headers=None,
             authenticator: Callable[[dict], dict] | None = 'default',
             response_processor: Callable[[Response], Any] | None = 'default'):
        return self.request('post', path, data, json, params, headers, authenticator, response_processor)

    def put(self, path, data=None, json=None, params=None, headers=None,
            authenticator: Callable[[dict], dict] | None = 'default',
            response_processor: Callable[[Response], Any] | None = 'default'):
        return self.request('put', path, data, json, params, headers, authenticator, response_processor)

    def patch(self, path, data=None, json=None, params=None, headers=None,
              authenticator: Callable[[dict], dict] | None = 'default',
              response_processor: Callable[[Response], Any] | None = 'default'):
        return self.request('patch', path, data, json, params, headers, authenticator, response_processor)
